#include <iostream>
#include <map>
#include <cstring>

using namespace std;

/** Product 1
* Represents the object under construction.
* Includes classes that define the parts. 
* Includes interfaces for assembling the parts.
*/
class CuddlyBear
{
	public:	
		CuddlyBear(string name = "Benny"){
			part["name"] = name;
		};
	
		CuddlyBear(const CuddlyBear& b) {
			part = b.part;
		}
		
		void setPart(string key, string value){
			part[key] = value;
		};

		void show() {
			cout<<"  NAME : "<< part["name"] << endl;
			cout<<"  BODY : "<< part["body"] << endl;
			cout<<"  STUFFING : "<< part["stuff"] << endl;
			cout<<"  HEART : "<< part["heart"] << endl;
			cout<<"  VOICE : "<< part["voice"] << endl;
			cout<<"---------------------------"<<endl;
		};

	private:
		map<string, string> part;
};

/** Product 2 */

class PlainGiraffe
{
	public:	
		PlainGiraffe(){
			strcpy(name,"Patches");
			strcpy(colour,"Brown");
			hight = 30;
			weight = 200;
		};

		PlainGiraffe(const PlainGiraffe& g) {
			strcpy(name, g.name);
			strcpy(colour, g.colour);
			hight = g.hight;
			weight= g.weight;
		}

		void  selectSkin(string skin = "Brown"){
			strcpy(colour, skin.c_str());
		};
		
		void setWeight(double w){
			weight = w;
		};
		
		void setHight(double h){
			hight = h;
		};
		
		void describe() {
			cout<<"  NAME : "<< name << endl;
			cout<<"  COLOUR : "<< colour << endl;
			cout<<"  WEIGHT : "<< weight << " gram" << endl;
			cout<<"  HEIGHT : " << hight  << " cm" << endl;
			cout<<"---------------------------"<<endl;
		};

	private:
		char name[10];
		char colour[20];
		double weight;
		double hight;
};


/** Builder
	Specifies an abstract interface for 
	creating parts of the Product object 
*/
class SoftToyMachine {
	public:
		virtual ~SoftToyMachine(){};
		virtual void selectBody(string s = "") = 0;
		virtual void stuff(string) = 0;
		virtual void insertHeart() {};
		virtual void addVoice() {};
		virtual void setName(string s =""){};
};

/** Director 1
	constructs an object using the Builder interface.

        This director receives its builder as a parameter
        of its constructor.
*/
class Shop {
	public:
		Shop(SoftToyMachine* m) : machine(m) , bearName("Teddy"){};
		virtual ~Shop(){ 
			/* note that the machine should not be deleted because it is
                            not owned by the shop - the client owns it */			
		};
		void construct() {
			machine->selectBody("best");
			machine->setName(bearName);
			machine->stuff("soft stuff");
			machine->insertHeart();
			machine->addVoice();
		};
	private:
		SoftToyMachine* machine;
		string bearName;
};

/** Director 2 
       notice that the process for constructing is different
       and does not contain all elements from the interface.

        This director provide a method to change
        its builder.
*/

class Mother {
	public:
		Mother() : machine(0) {};
		virtual ~Mother(){};

		void construct() {
			machine->selectBody();
			machine->insertHeart();
			machine->stuff("medium stuff");
		};

		void setMachine(SoftToyMachine* m) {
			machine = m;
		};
	private:
		SoftToyMachine* machine;
};

/** ConcreteBuilder 1
* builds the product's internal representation
* defines the process by which it is assembled.
* defines and keeps track of the representation it creates
* provides an interface for retrieving the product
*/

class CuddlyBearMaker : public SoftToyMachine {
	public:
		CuddlyBearMaker() {}
    
		void selectBody(string category) {
			bear.setPart("body","Bear");
		}

		void stuff(string stuffing) {
			bear.setPart("stuff",stuffing);
		}

		void insertHeart() {
			bear.setPart("heart","Large red");
		}

		void addVoice() {
			bear.setPart("voice","Hello");
		}

		void setName(string name){
			bear.setPart("name", name);
		}
		
		CuddlyBear getBear(){
			return bear;
		}
		
	private:
		CuddlyBear bear;
};


/** ConcreteBuilder 2
*/
class PlainGiraffeMaker :public SoftToyMachine {
	public:
		PlainGiraffeMaker() {};
		virtual ~PlainGiraffeMaker(){
				if (giraffe) delete giraffe;
		};
		
		void selectBody(string colour) {
			if (giraffe) delete giraffe;
			giraffe = new PlainGiraffe();
			if(colour.compare("best") == 0)
				giraffe->selectSkin("Giraffe pattern");
		};
		
		void stuff(string stuffing) {
			if (stuffing.compare("soft stuff") == 0)
				giraffe->setWeight(100);
		} ; 
   
		PlainGiraffe getGiraffe(){
			PlainGiraffe* replica = new PlainGiraffe(*giraffe);
			return  *replica;
		};

	private:
		PlainGiraffe* giraffe;
};

/** Client
* Creates Directors and  Concrete Builders 
* Use the Director to construct a product
* Retrieve the final product from the Concrete Builder
*/

int main(){
  /*Variable to store Products */ 
 CuddlyBear bear[2];
 PlainGiraffe giraffe[2];
	
  /*Concrete builders  */
   SoftToyMachine* machine0 = new PlainGiraffeMaker();
   SoftToyMachine* machine1 = new CuddlyBearMaker();
   SoftToyMachine* machine2 = new CuddlyBearMaker();
	
/*Director: Shop uses the same builders to construct better variants of these toys */
  Shop* shop = new Shop(machine0);	
   shop->construct();
   giraffe[1] = ((PlainGiraffeMaker*) machine0)->getGiraffe(); 
   delete shop;
   shop = new Shop(machine1);
   shop->construct();	
   bear[1] = ((CuddlyBearMaker*) machine1)->getBear();

  /*Director : Mother  creates one of each toy type mainly sticking to defaults 
	using lower quality stuff and leaving out some of the finer touches */
  Mother* me = new Mother();
  me->setMachine(machine2);
  me->construct();
  bear[0] = ((CuddlyBearMaker*) machine2)->getBear();
  me->setMachine(machine0);
  me->construct();
  giraffe[0] = ((PlainGiraffeMaker*) machine0)->getGiraffe();	

 /*The products */
  cout << "\n ============ MOM: ============\n";
  bear[0].show();
  giraffe[0].describe();
  cout << "\n ============ SHOP: ===========\n";
  bear[1].show();
  giraffe[1].describe();
  
  /* clean up */
delete me;
delete  shop;
delete machine0;
delete machine1;
  
  return 0;
}
